/*------------------------------------------------------------------------
 * Implementation of functions defined in portable.h for the ARM CM3 port.
 *
 * Stevan A. Milinkovic
 * Union University, School of Computing
 * Knez Mihailova 6, 11000 Belgrade, Serbia
 * e-mail: smilinkovic@raf.edu.rs
 *-----------------------------------------------------------------------*/

/* Scheduler includes. */
#include "FreeRTOS.h"
#include "task.h"

#ifndef configKERNEL_INTERRUPT_PRIORITY
        #define configKERNEL_INTERRUPT_PRIORITY 255
#endif

/* Constants required to manipulate the NVIC. */
#define portNVIC_SYSTICK_CTRL          ( ( volatile unsigned long *) 0xe000e010 )
#define portNVIC_SYSTICK_LOAD          ( ( volatile unsigned long *) 0xe000e014 )
#define portNVIC_INT_CTRL              ( ( volatile unsigned long *) 0xe000ed04 )
#define portNVIC_SYSPRI2               ( ( volatile unsigned long *) 0xe000ed20 )
#define portNVIC_SYSTICK_CLK           0x00000004
#define portNVIC_SYSTICK_INT           0x00000002
#define portNVIC_SYSTICK_ENABLE        0x00000001
#define portNVIC_PENDSVSET             0x10000000
#define portNVIC_PENDSV_PRI            ( ( ( unsigned long ) configKERNEL_INTERRUPT_PRIORITY ) << 16 )
#define portNVIC_SYSTICK_PRI           ( ( ( unsigned long ) configKERNEL_INTERRUPT_PRIORITY ) << 24 )

/* Constants required to set up the initial stack. */
#define portINITIAL_XPSR               ( 0x01000000 )

#define basepri     17
#define msp         8
#define psp         9

extern pxCurrentTCB;

/* Each task maintains its own interrupt status in the critical nesting variable. */
static unsigned portBASE_TYPE uxCriticalNesting = 0xaaaaaaaa;

/* Setup the timer to generate the tick interrupts.*/
static void prvSetupTimerInterrupt( void );

/* Exception handlers.*/
void xPortPendSVHandler( void );
void xPortSysTickHandler( void );
void vPortSVCHandler( void );

/* Start first task is a separate function so it can be tested in isolation. */
void vPortStartFirstTask( void );

/* Simulate the stack frame as it would be created by a context switch interrupt. */
portSTACK_TYPE *pxPortInitialiseStack( portSTACK_TYPE *pxTopOfStack, pdTASK_CODE pxCode, void *pvParameters ) {
        pxTopOfStack--;  /* Offset added to account for the way the MCU uses the stack on entry/exit of interrupts. */
        *pxTopOfStack = portINITIAL_XPSR;             /* xPSR */
        pxTopOfStack--;
        *pxTopOfStack = ( portSTACK_TYPE ) pxCode;    /* PC */
        pxTopOfStack--;
        *pxTopOfStack = 0;        /* LR */
        pxTopOfStack -= 5;        /* R12, R3, R2 and R1. */
        *pxTopOfStack = ( portSTACK_TYPE ) pvParameters;        /* R0 */
        pxTopOfStack -= 8;        /* R11, R10, R9, R8, R7, R6, R5 and R4. */

        return pxTopOfStack;
}
/*-----------------------------------------------------------*/

portBASE_TYPE xPortStartScheduler( void ) {
        /* Make PendSV, CallSV and SysTick the same priroity as the kernel. */
        *(portNVIC_SYSPRI2) |= portNVIC_PENDSV_PRI;
        *(portNVIC_SYSPRI2) |= portNVIC_SYSTICK_PRI;

        /* Start the timer that generates the tick ISR.  Interrupts are disabled here already. */
        prvSetupTimerInterrupt();
        
        /* Initialise the critical nesting count ready for the first task. */
        uxCriticalNesting = 0;

        /* Start the first task. */
        vPortStartFirstTask();

        /* Should not get here! */
        return 0;
}
/*-----------------------------------------------------------*/

void vPortEndScheduler( void ) {
        /* It is unlikely that the CM3 port will require this function as there is nothing to return to.  */
}
/*-----------------------------------------------------------*/

void vPortYieldFromISR( void ) {
        /* Set a PendSV to request a context switch. */
        *(portNVIC_INT_CTRL) = portNVIC_PENDSVSET;
}
/*-----------------------------------------------------------*/

void vPortEnterCritical( void ) {
        portDISABLE_INTERRUPTS();
        uxCriticalNesting++;
}
/*-----------------------------------------------------------*/

void vPortExitCritical( void ) {
        uxCriticalNesting--;
        if( uxCriticalNesting == 0 ) {
                portENABLE_INTERRUPTS();
        }
}
/*-----------------------------------------------------------*/

void xPortSysTickHandler(void) iv IVT_INT_SysTick ics ICS_OFF {
  unsigned long ulDummy;
        /* If using preemption, also force a context switch. */
        #if configUSE_PREEMPTION == 1
                *(portNVIC_INT_CTRL) = portNVIC_PENDSVSET;        
        #endif

        ulDummy = portSET_INTERRUPT_MASK_FROM_ISR();
        vTaskIncrementTick();
        portCLEAR_INTERRUPT_MASK_FROM_ISR( ulDummy );
}
/*-----------------------------------------------------------*/

/* Setup the systick timer to generate the tick interrupts at the required frequency. */
static void prvSetupTimerInterrupt( void ) {
        /* Configure SysTick to interrupt at the requested rate. */
        *(portNVIC_SYSTICK_LOAD) = ( configCPU_CLOCK_HZ / configTICK_RATE_HZ ) - 1UL;
        *(portNVIC_SYSTICK_CTRL) = portNVIC_SYSTICK_CLK | portNVIC_SYSTICK_INT | portNVIC_SYSTICK_ENABLE;
}



/*-----------------------------------------------------------*/
void vPortSVCHandler(void) iv IVT_INT_SVCall ics ICS_OFF {
__asm {
        movw  r3, #lo_addr(_pxCurrentTCB)
        movt  r3, #hi_addr(_pxCurrentTCB)
        ldr   r1, [r3]                  /* Get the pxCurrentTCB address. */
        ldr   r0, [r1]                  /* The first item in pxCurrentTCB is the task top of stack. */
        ldm   r0!, (r4-r11)             /* Pop the registers that are not automatically saved on exception entry and the critical nesting count. */
        msr   psp, r0                   /* Restore the task stack pointer. */
        mov   r0, #0
        msr   basepri, r0
        orr   r14, #0xd                 /* Return to thread mode and on return use the process stack (LR = 0xFFFFFFFD) */
      }
}

/*-----------------------------------------------------------*/
void vPortStartFirstTask( void ) {
__asm {
        movw  r0, #0xED08               /* Use the NVIC offset register to locate the stack. */
        movt  r0, #0xE000
        ldr   r0, [r0]
        ldr   r0, [r0]
        msr   msp, r0                   /* Set the msp back to the start of the stack. */
        cpsie i                         /* Globally enable interrupts. */
        svc   #0                        /* SVC to start the first task. */
      }
}

/*-----------------------------------------------------------*/
void xPortPendSVHandler (void) iv IVT_INT_PendSV ics ICS_OFF {
__asm {
        mrs   r0, psp                   /* Get the location of the current TCB. */
        movw  r3, #lo_addr(_pxCurrentTCB)
        movt  r3, #hi_addr(_pxCurrentTCB)
        ldr   r2, [r3]
        stmdb r0!, (r4-r11)             /* Save the remaining registers. */
        str   r0, [r2]                  /* Save the new top of stack into the first member of the TCB. */
        stmdb sp!, (r3, r14)
        mov   r0, # configMAX_SYSCALL_INTERRUPT_PRIORITY
        msr   basepri, r0
        bl    _vTaskSwitchContext       /* Call scheduler  */
        mov   r0, #0
        msr   basepri, r0
        ldm   sp!, (r3, r14)
        ldr   r1, [r3]
        ldr   r0, [r1]                  /* The first item in pxCurrentTCB is the task top of stack. */
        ldm   r0!, (r4-r11)             /* Pop the registers and the critical nesting count. */
        msr   psp, r0
      }
 }

/*-----------------------------------------------------------*/
void vPortSetInterruptMask( void ) {
__asm {
        push  (r0)
        mov   r0, # configMAX_SYSCALL_INTERRUPT_PRIORITY
        msr   basepri, r0
        pop   (r0)
      }
}

/*-----------------------------------------------------------*/
void vPortClearInterruptMask( void ) {
__asm {
        push   (r0)
        mov    r0, #0
        msr    basepri, r0
        pop    (r0)
      }
}